from datetime import datetime

from pydantic import BaseModel, Field, AliasChoices, field_validator, field_serializer

from ..jupiter.param import JupiterSwapMode, JupiterSwapDex, JupiterLimitOrderState

# class used on Jupiter HTTPErrors
class JupiterHTTPError(BaseModel):
    """
        Jupiter API returns an error schema on failed request 
        that can be used to investigated the error. This schema 
        is used to strandardise the HTTPErrors.
    """
    code: str = Field(validation_alias = AliasChoices("errorCode", "error_code"))
    msg: str = Field(alias = "error")

# classes used on GET "Price" endpoint
class GetPriceData(BaseModel):
    id: str
    mint_symbol: str = Field(alias = "mintSymbol")
    vs_token: str = Field(alias = "vsToken")
    vs_token_symbol: str = Field(alias = "vsTokenSymbol")
    price: float

class GetPriceResponse(BaseModel):
    """
        Model used to represent the **Price** endpoint from Jupiter API.
    """
    data: dict[str, GetPriceData]
    time_taken: float = Field(alias = "timeTaken")

# classes used on GET "Quote" endpoint
# Input
class GetQuoteInput(BaseModel):
    """
        Model used to identify the inputs params required by 
        a GET Quote request.
    """

    input_token: str = Field(serialization_alias = "inputMint")
    """The address of the input token on the chain."""

    output_token: str = Field(serialization_alias = "outputMint")
    """The address of the output token on the chain."""

    amount: int
    """The amount to swap, factoring in the token decimals."""

    slippage_base_points: int = Field(default = 50, serialization_alias = "slippageBps")
    """Slippage tolerance in basis points (the default is used unless `auto_slippage` is set to `True`).  
        Observe that if the slippage exeeded this value, then the swap will fail."""

    swap_mode: str = Field(default = JupiterSwapMode.EXACT_IN.value, serialization_alias = "swapMode")
    """Define if the slippage is on the input or output token."""

    dexes: list[str] | None = None
    """List of DEXes to include; by default, all the DEXes are included.  
        See [`JupiterSwapDex`][cyhole.jupiter.param.JupiterSwapDex] for all the supported DEXs"""

    exclude_dexes: list[str] | None = Field(default = None, serialization_alias = "excludeDexes")
    """List of DEXes to exclude.  
        See [`JupiterSwapDex`][cyhole.jupiter.param.JupiterSwapDex] for all the supported DEXs"""

    restrict_intermediate_tokens: bool | None = Field(default = None, serialization_alias = "restrictIntermediateTokens")
    """Restrict to a top token set for stable liquidity."""

    only_direct_routes: bool = Field(default = False, serialization_alias = "onlyDirectRoutes")
    """Limit to single hop routes only."""

    as_legacy_transaction: bool = Field(default = False, serialization_alias = "asLegacyTransaction")
    """Use legacy transactions instead of versioned ones."""

    platform_fee_base_points: int | None = Field(default = None, serialization_alias = "platformFeeBps")
    """Fee to charge. The value is in percent and taken from output token."""

    max_accounts: int | None = Field(default = None, serialization_alias = "maxAccounts")
    """Max accounts to be used for the quote."""

    auto_slippage: bool = Field(default = False, serialization_alias = "autoSlippage")
    """Enable smart/auto slippage proposed by API."""

    max_auto_slippage_base_points: int | None = Field(default = None, serialization_alias = "maxAutoSlippageBps")
    """Max slippage for smart/auto slippage."""

    auto_slippage_collision_usd_value: int | None = Field(default = None, serialization_alias = "autoSlippageCollisionUsdValue")
    """Custom USD value for calculating slippage impact.  
        By default, the API sets 1000 USD if `auto_slippage` is set to `True`."""

    @field_validator("dexes", "exclude_dexes")
    @classmethod
    def validator_dexes(cls, dexes: list[str]) -> list[str]:
        for dex in dexes:
            JupiterSwapDex.check(dex)
        return dexes

    @field_serializer("dexes", "exclude_dexes")
    @classmethod
    def serialize_dexes(cls, dexes: list[str] | None) -> str | None:
        return ",".join(dexes) if dexes else None

    @field_validator("swap_mode")
    @classmethod
    def validator_swap_mode(cls, mode: str) -> str:
        JupiterSwapMode.check(mode)
        return mode

# Output
class GetQuotePlatformFees(BaseModel):
    amount: str
    fee_base_points: str = Field(alias = "feeBps")

class GetQuoteSwapInfo(BaseModel):
    amm_key: str = Field(alias = "ammKey")
    amm_label: str | None = Field(default = None, alias = "label")
    input_token: str = Field(alias = "inputMint")
    input_amount: str = Field(alias = "inAmount")
    output_token: str = Field(alias = "outputMint")
    output_amount: str = Field(alias = "outAmount")
    fee_token: str = Field(alias = "feeMint")
    fee_amount: str = Field(alias = "feeAmount")

class GetQuoteRoutePlan(BaseModel):
    swap_info: GetQuoteSwapInfo = Field(alias = "swapInfo")
    percent: int

class GetQuoteResponse(BaseModel):
    """
        Model used to represent the **Quote** endpoint from Jupiter API.
    """
    input_token: str = Field(alias = "inputMint")
    input_amount: str = Field(alias = "inAmount")
    output_token: str = Field(alias = "outputMint")
    output_amount: str = Field(alias = "outAmount")
    other_amount_threshold: str = Field(alias = "otherAmountThreshold")
    swap_mode: str = Field(alias = "swapMode")
    slippage_base_points: int = Field(alias = "slippageBps")
    platform_fees: GetQuotePlatformFees | None = Field(default = None, alias = "platformFee")
    price_impact_pct: str = Field(alias = "priceImpactPct")
    route_plan: list[GetQuoteRoutePlan] = Field(alias = "routePlan")
    context_slot: int = Field(alias = "contextSlot")
    time_taken: float = Field(alias = "timeTaken")

# classes used on GET "Quote/Token" endpoint
class GetQuoteTokensResponse(BaseModel):
    """
        Model used to represent the **Quote/Token** endpoint from Jupiter API.
    """
    tokens: list[str]

# classes used on GET "Quote/Program ID to Label" endpoint
class GetQuoteProgramIdLabelResponse(BaseModel):
    """
        Model used to represent the **Quote/Program ID to Label** endpoint from Jupiter API.
    """
    dexes: dict[str, str]

# classes used on POST "Swap" endpoint
# Body
class PostSwapBody(BaseModel):
    """
        Model used to identify the body required by a POST **Swap** request.
    """
    user_public_key: str = Field(serialization_alias = "userPublicKey")
    wrap_unwrap_sol: bool = Field(default = None, serialization_alias = "wrapAndUnwrapSol")
    use_shared_accounts: bool= Field(default = None, serialization_alias = "useSharedAccounts")
    fee_account: str = Field(default = None, serialization_alias = "feeAccount")
    tracking_account: str = Field(default = None, serialization_alias = "trackingAccount")
    compute_unit_price_micro_lamports: int = Field(default = None, serialization_alias = "computeUnitPriceMicroLamports")
    prioritization_fee_lamports: int = Field(default = None, serialization_alias = "prioritizationFeeLamports")
    as_legacy_transaction: bool = Field(default = None, serialization_alias = "asLegacyTransaction")
    use_token_ledger: bool = Field(default = None, serialization_alias = "useTokenLedger")
    destination_token_account: str = Field(default = None, serialization_alias = "destinationTokenAccount")
    dynamic_compute_unit_limit: bool = Field(default = None, serialization_alias = "dynamicComputeUnitLimit")
    skip_user_accounts_rpc_calls: bool = Field(default = None, serialization_alias = "skipUserAccountsRpcCalls")
    quote_response: GetQuoteResponse = Field(serialization_alias = "quoteResponse")

# Output
class PostSwapResponse(BaseModel):
    swap_transaction: str = Field(alias = "swapTransaction")
    last_valid_block_height: int = Field(alias = "lastValidBlockHeight")
    prioritization_fee_lamports: int = Field(default = 0, alias = "prioritizationFeeLamports")

# classes used on GET "Token List" endpoint
class GetTokenListToken(BaseModel):
    address: str
    chain_id: int = Field(alias = "chainId")
    decimals: int
    name: str
    symbol: str
    logo_uri: str = Field(alias = "logoURI")
    tags: list[str]
    extensions: dict[str, str] | None = None

class GetTokenListResponse(BaseModel):
    """
        Model used to represent the **Token List** endpoint from Jupiter API.
    """
    tokens: list[GetTokenListToken]

# classes used on POST "Limit Order Create" endpoint
# Body
class PostLimitOrderCreateBody(BaseModel):
    """
        Model used to identify the body required by a POST **Limit Order Create** request.
    """
    user_public_key: str = Field(serialization_alias = "owner")
    """Public Key of the Owner wallet"""

    input_token: str = Field(serialization_alias = "inputMint")
    """The address of the input token on the chain used to buy."""

    input_amount: int = Field(serialization_alias = "inAmount")
    """The amount of input token to use for the limit order."""

    output_token: str = Field(serialization_alias = "outputMint")
    """The address of the output token on the chain that will bought."""

    output_amount: int = Field(serialization_alias = "outAmount")
    """The amount of output token to buy in the limit order."""

    base: str
    """Public Key used to initiate the Limit Order"""

    expired_at_unix_time: int | None = Field(default = None, serialization_alias = "expiredAt")
    """Expiring date for the Limit Order expressed in UNIX time"""

    referral_public_key: str | None = Field(default = None, serialization_alias = "referralAccount")
    """The address of the account used to get referral fees."""

    referral_name: str | None = Field(default = None, serialization_alias = "referralName")

# Output
class PostLimitOrderCreateResponse(BaseModel):
    """
        Model used to represent the **Limit Order Create** endpoint from Jupiter API.
    """
    transaction: str = Field(alias = "tx")
    order_public_key: str = Field(alias = "orderPubkey")

# classes used on POST "Limit Order Cancel" endpoint
# Body
class PostLimitOrderCancelBody(BaseModel):
    """
        Model used to identify the body required by a POST **Limit Order Cancel** request.
    """
    user_public_key: str = Field(serialization_alias = "owner")
    """Public Key of the Owner wallet"""

    fee_payer_public_key: str = Field(serialization_alias = "feePayer")
    """Public Key of the fee payer."""

    orders: list[str]
    """List of orders Public Keys to cancel."""

# Output
class PostLimitOrderCancelResponse(BaseModel):
    """
        Model used to represent the **Limit Order Cancel** endpoint from Jupiter API.
    """
    transaction: str = Field(alias = "tx")

# classes used on GET "Limit Order Opens" endpoint
class GetLimitOrderOpenAccount(BaseModel):
    maker: str
    input_token: str = Field(alias = "inputMint")
    input_amount: str = Field(alias = "inAmount")
    output_token: str = Field(alias = "outputMint")
    output_amount: str = Field(alias = "outAmount")
    ori_input_token: str = Field(alias = "oriInAmount")
    ori_output_amount: str = Field(alias = "oriOutAmount")
    expired_at_unix_time: int | None = Field(default = None, alias = "expiredAt")
    base: str

class GetLimitOrderOpen(BaseModel):
    public_key: str = Field(alias = "publicKey")
    account: GetLimitOrderOpenAccount

class GetLimitOrderOpenResponse(BaseModel):
    """
        Model used to represent the **Limit Order Opens** endpoint from Jupiter API.
    """
    orders: list[GetLimitOrderOpen]

# classes used on GET "Limit Order History" endpoint
class GetLimitOrderHistory(BaseModel):
    id: int
    maker: str
    order_key: str = Field(alias = "orderKey")
    input_token: str = Field(alias = "inputMint")
    input_amount: str = Field(alias = "inAmount")
    output_token: str = Field(alias = "outputMint")
    output_amount: str = Field(alias = "outAmount")
    ori_input_token: str = Field(alias = "oriInAmount")
    ori_output_amount: str = Field(alias = "oriOutAmount")
    expired_at_unix_time: int | None = Field(default = None, alias = "expiredAt")
    state: str
    create_transaction_id: str = Field(alias = "createTxid")
    cancel_transaction_id: str | None = Field(default = None, alias = "cancelTxid")
    updated_at: datetime = Field(alias = "updatedAt")
    created_at: datetime = Field(alias = "createdAt")

    @field_validator("created_at", "updated_at")
    def parse_datetime(cls, datetime_raw: str | datetime) -> datetime:
        if isinstance(datetime_raw, str):
            return datetime.strptime(datetime_raw, "%Y-%m-%dT%H:%M:%S")
        return datetime_raw

    @field_validator("state")
    @classmethod
    def validator_state(cls, state: str) -> str:
        JupiterLimitOrderState.check(state)
        return state

class GetLimitOrderHistoryResponse(BaseModel):
    """
        Model used to represent the **Limit Order History** endpoint from Jupiter API.
    """
    orders: list[GetLimitOrderHistory]

# classes used on GET "Limit Order Trade History" endpoint
class GetLimitOrderTradeHistoryOrder(BaseModel):
    id: int
    order_key: str = Field(alias = "orderKey")
    input_token: str = Field(alias = "inputMint")
    output_token: str = Field(alias = "outputMint")

class GetLimitOrderTradeHistory(BaseModel):
    id: int
    input_amount: str = Field(alias = "inAmount")
    output_amount: str = Field(alias = "outAmount")
    transaction_id: str = Field(alias = "txid")
    updated_at: datetime = Field(alias = "updatedAt")
    created_at: datetime = Field(alias = "createdAt")
    order: GetLimitOrderTradeHistoryOrder

    @field_validator("created_at", "updated_at")
    def parse_datetime(cls, datetime_raw: str | datetime) -> datetime:
        if isinstance(datetime_raw, str):
            return datetime.strptime(datetime_raw, "%Y-%m-%dT%H:%M:%S")
        return datetime_raw

class GetLimitOrderTradeHistoryResponse(BaseModel):
    """
        Model used to represent the **Limit Order Trade History** endpoint from Jupiter API.
    """
    orders: list[GetLimitOrderTradeHistory]


import pytest
from pathlib import Path

from pytest_mock import MockerFixture

from cyhole.jupiter import Jupiter
from cyhole.jupiter.schema import (
    GetPriceResponse,
    GetQuoteInput,
    GetQuoteResponse,
    GetQuoteTokensResponse,
    GetQuoteProgramIdLabelResponse,
    PostSwapBody,
    PostSwapResponse,
    GetTokenListResponse,
    PostLimitOrderCreateBody,
    PostLimitOrderCreateResponse,
    PostLimitOrderCancelBody,
    PostLimitOrderCancelResponse,
    GetLimitOrderOpenResponse,
    GetLimitOrderHistoryResponse,
    GetLimitOrderTradeHistoryResponse
)
from cyhole.jupiter.param import JupiterSwapDex, JupiterSwapMode
from cyhole.jupiter.exception import JupiterNoRouteFoundError, JupiterException
from cyhole.core.address.solana import SOL, JUP, USDC, BONK
from cyhole.core.address.ethereum import WETH
from cyhole.core.exception import ParamUnknownError

# load test config
from .config import load_config, MockerManager
config = load_config()

# create resources folder
mock_path = Path(config.mock_folder) / config.jupiter.mock_folder
mock_path.mkdir(parents = True, exist_ok = True)

class TestJupiter:
    """
        Class grouping all unit tests.
    """
    jupiter = Jupiter()
    mocker = MockerManager(mock_path)

    def test_get_price_token_address_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic. Only one token address.

            Mock Response File: get_price_token_address.json
        """

        # load mock response
        mock_file_name = "get_price_token_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.jupiter.client.get_price([JUP])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert JUP in response.data

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_token_address_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic. Only one token address.

            Mock Response File: get_price_token_address.json
        """

        # load mock response
        mock_file_name = "get_price_token_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_price([JUP])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert JUP in response.data

    def test_get_price_token_symbol_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic. Only one token symbol.

            Mock Response File: get_price_token_symbol.json
        """

        token_symbol = "JUP"
        # load mock response
        mock_file_name = "get_price_token_symbol"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.jupiter.client.get_price([token_symbol])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert token_symbol in response.data

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_token_symbol_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic. Only one token symbol.

            Mock Response File: get_price_token_symbol.json
        """

        token_symbol = "JUP"
        # load mock response
        mock_file_name = "get_price_token_symbol"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_price([token_symbol])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert token_symbol in response.data

    def test_get_price_multiple_token_address_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic. More then one token addresses.

            Mock Response File: get_price_multiple_token_address.json
        """

        # load mock response
        mock_file_name = "get_price_multiple_token_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.jupiter.client.get_price([JUP, SOL])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert (JUP in response.data) and (SOL in response.data)

    @pytest.mark.asyncio
    async def test_get_price_multiple_token_address_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic. More then one token addresses.

            Mock Response File: get_price_multiple_token_address.json
        """

        # load mock response
        mock_file_name = "get_price_multiple_token_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_price([JUP, SOL])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert (JUP in response.data) and (SOL in response.data)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    def test_get_price_vs_address_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic. Provide a different comparison token.

            Mock Response File: get_price_vs_address.json
        """

        # load mock response
        mock_file_name = "get_price_vs_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.jupiter.client.get_price([SOL], vs_address = JUP)

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert SOL in response.data
        assert response.data[SOL].vs_token == JUP

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_vs_address_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic. Provide a different comparison token.

            Mock Response File: get_price_vs_address.json
        """

        # load mock response
        mock_file_name = "get_price_vs_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_price([SOL], vs_address = JUP)

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert SOL in response.data
        assert response.data[SOL].vs_token == JUP

    def test_get_price_unknown_address_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic. Provide an unknown token address.

            Mock Response File: get_price_unknown_address.json
        """

        # load mock response
        mock_file_name = "get_price_unknown_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.jupiter.client.get_price([WETH])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert response.data == {}

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_unknown_address_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic. Provide an unknown token address.

            Mock Response File: get_price_unknown_address.json
        """

        # load mock response
        mock_file_name = "get_price_unknown_address"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_price([WETH])

        # actual test
        assert isinstance(response, GetPriceResponse)
        assert response.data == {}

    def test_get_quote_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" for synchronous logic.

            Mock Response File: get_quote_base.json
        """

        # load mock response
        mock_file_name = "get_quote_base"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        amount = 1000
        # execute request
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = amount
        )
        response = self.jupiter.client.get_quote(input)

        # actual test
        assert isinstance(response, GetQuoteResponse)
        assert response.input_amount == str(amount)
        assert response.input_token == SOL
        assert response.output_token == JUP

    @pytest.mark.asyncio
    async def test_get_quote_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" for asynchronous logic.

            Mock Response File: get_quote_base.json
        """

        # load mock response
        mock_file_name = "get_quote_base"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        amount = 1000
        # execute request
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = amount
        )
        async with self.jupiter.async_client as client:
            response = await client.get_quote(input)

        # actual test
        assert isinstance(response, GetQuoteResponse)
        assert response.input_amount == str(amount)
        assert response.input_token == SOL
        assert response.output_token == JUP

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    def test_get_quote_force_route_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            forcing a route and mode for synchronous logic.

            Mock Response File: get_quote_force_rooute.json
        """

        # load mock response
        mock_file_name = "get_quote_force_rooute"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        amount = 1000
        # execute request
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = amount,
            dexes = [JupiterSwapDex.METEORA_DLMM.value],
            swap_mode = JupiterSwapMode.EXACT_IN.value
        )
        response = self.jupiter.client.get_quote(input)

        # actual test
        assert isinstance(response, GetQuoteResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_quote_force_route_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            forcing a route and mode for asynchronous logic.

            Mock Response File: get_quote_force_rooute.json
        """

        # load mock response
        mock_file_name = "get_quote_force_rooute"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        amount = 1000
        # execute request
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = amount,
            dexes = [JupiterSwapDex.METEORA_DLMM.value],
            swap_mode = JupiterSwapMode.EXACT_IN.value
        )

        async with self.jupiter.async_client as client:
            response = await client.get_quote(input)

        # actual test
        assert isinstance(response, GetQuoteResponse)

    def test_get_quote_error_route_not_found_sync(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            when no route is found for synchronous logic.
        """

        # define input
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = 1
        )

        # actual test
        with pytest.raises(JupiterNoRouteFoundError):
            self.jupiter.client.get_quote(input)

    @pytest.mark.asyncio
    async def test_get_quote_error_route_not_found_async(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            when no route is found for asynchronous logic.
        """

        # define input
        input = GetQuoteInput(
            input_token = SOL,
            output_token = JUP,
            amount = 1
        )

        # actual test
        with pytest.raises(JupiterNoRouteFoundError):
            async with self.jupiter.async_client as client:
                await client.get_quote(input)

    def test_get_quote_error_unknown_dex(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            when a not supported DEX is used.
        """

        # actual test
        with pytest.raises(ParamUnknownError):
            GetQuoteInput(
                input_token = SOL,
                output_token = JUP,
                amount = 1000,
                dexes = ["XXX"]
            )

    def test_get_quote_error_unknown_mode(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote" 
            when a not supported MODE is used.
        """

        # actual test
        with pytest.raises(ParamUnknownError):
            GetQuoteInput(
                input_token = SOL,
                output_token = JUP,
                amount = 1000,
                swap_mode = "XXX"
            )

    def test_get_quote_tokens_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote/Tokens" for synchronous logic.

            Mock Response File: get_quote_tokens.json
        """

        # load mock response
        mock_file_name = "get_quote_tokens"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteTokensResponse)

            # response content to be adjusted
            content = str(mock_response.json()["tokens"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_quote_tokens()

        # actual test
        assert isinstance(response, GetQuoteTokensResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_quote_tokens_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote/Tokens" for asynchronous logic.

            Mock Response File: get_quote_tokens.json
        """

        # load mock response
        mock_file_name = "get_quote_tokens"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteTokensResponse)

            # response content to be adjusted
            content = str(mock_response.json()["tokens"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_quote_tokens()

        # actual test
        assert isinstance(response, GetQuoteTokensResponse)

    def test_get_quote_program_id_label_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote/Program ID to Label" 
            for synchronous logic.

            Mock Response File: get_quote_program_id_label.json
        """

        # load mock response
        mock_file_name = "get_quote_program_id_label"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteProgramIdLabelResponse)

            # response content to be adjusted
            content = str(mock_response.json()["dexes"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_quote_program_id_label()

        # actual test
        assert isinstance(response, GetQuoteProgramIdLabelResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_quote_program_id_label_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Quote/Program ID to Label" 
            for asynchronous logic.

            Mock Response File: get_quote_program_id_label.json
        """

        # load mock response
        mock_file_name = "get_quote_program_id_label"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetQuoteProgramIdLabelResponse)

            # response content to be adjusted
            content = str(mock_response.json()["dexes"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_quote_program_id_label()

        # actual test
        assert isinstance(response, GetQuoteProgramIdLabelResponse)

    def test_post_swap_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Swap" 
            for synchronous logic.

            Mock Response File: post_swap.json
        """

        # load mock response
        mock_file_name = "post_swap"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, PostSwapResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        body = PostSwapBody(
            user_public_key = "REFER4ZgmyYx9c6He5XfaTMiGfdLwRnkV4RPp9t9iF3",
            quote_response = self.mocker.load_mock_model("get_quote_base", GetQuoteResponse)
        )
        response = self.jupiter.client.post_swap(body)

        # actual test
        assert isinstance(response, PostSwapResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_post_swap_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Swap" 
            for asynchronous logic.

            Mock Response File: post_swap.json
        """

        # load mock response
        mock_file_name = "post_swap"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, PostSwapResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        body = PostSwapBody(
            user_public_key = "REFER4ZgmyYx9c6He5XfaTMiGfdLwRnkV4RPp9t9iF3",
            quote_response = self.mocker.load_mock_model("get_quote_base", GetQuoteResponse)
        )
        async with self.jupiter.async_client as client:
            response = await client.post_swap(body)

        # actual test
        assert isinstance(response, PostSwapResponse)

    def test_post_swap_invalid_request_sync(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Swap" 
            when an invalid field is provided in the body for synchronous logic.
        """
        body = PostSwapBody(
            user_public_key = "XXX",
            quote_response = self.mocker.load_mock_model("get_quote_base", GetQuoteResponse)
        )
        with pytest.raises(JupiterException):
            self.jupiter.client.post_swap(body)

    @pytest.mark.asyncio
    async def test_post_swap_invalid_request_async(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Swap" 
            when an invalid field is provided in the body for asynchronous logic.
        """
        body = PostSwapBody(
            user_public_key = "XXX",
            quote_response = self.mocker.load_mock_model("get_quote_base", GetQuoteResponse)
        )
        with pytest.raises(JupiterException):
            async with self.jupiter.async_client as client:
                await client.post_swap(body)

    def test_get_token_list_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token List" 
            for synchronous logic.

            Mock Response File: get_token_list.json
        """

        # load mock response
        mock_file_name = "get_token_list"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenListResponse)

            # response content to be adjusted
            content = str(mock_response.json()["tokens"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_token_list()

        # actual test
        assert isinstance(response, GetTokenListResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            response.tokens = response.tokens[0:10]
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_list_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token List" 
            for asynchronous logic.

            Mock Response File: get_token_list.json
        """

        # load mock response
        mock_file_name = "get_token_list"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenListResponse)

            # response content to be adjusted
            content = str(mock_response.json()["tokens"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_token_list()

        # actual test
        assert isinstance(response, GetTokenListResponse)

    def test_post_limit_order_create_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Limit Order Create" 
            for synchronous logic.

            Mock Response File: post_limit_order_create.json
        """

        # load mock response
        mock_file_name = "post_limit_order_create"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, PostLimitOrderCreateResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        body = PostLimitOrderCreateBody(
            user_public_key = "REFER4ZgmyYx9c6He5XfaTMiGfdLwRnkV4RPp9t9iF3",
            input_token = USDC,
            input_amount = 100_000,
            output_token = JUP,
            output_amount = 100_000,
            base = "5pVyoAeURQHNMVU7DmfMHvCDNmTEYXWfEwc136GYhTKG"
        )
        response = self.jupiter.client.post_limit_order_create(body)

        # actual test
        assert isinstance(response, PostLimitOrderCreateResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_post_limit_order_create_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Limit Order Create" 
            for asynchronous logic.

            Mock Response File: post_limit_order_create.json
        """

        # load mock response
        mock_file_name = "post_limit_order_create"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, PostLimitOrderCreateResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        body = PostLimitOrderCreateBody(
            user_public_key = "REFER4ZgmyYx9c6He5XfaTMiGfdLwRnkV4RPp9t9iF3",
            input_token = USDC,
            input_amount = 100_000,
            output_token = JUP,
            output_amount = 100_000,
            base = "5pVyoAeURQHNMVU7DmfMHvCDNmTEYXWfEwc136GYhTKG"
        )
        async with self.jupiter.async_client as client:
            response = await client.post_limit_order_create(body)

        # actual test
        assert isinstance(response, PostLimitOrderCreateResponse)

    def test_post_limit_order_cancel_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Limit Order Cancel" 
            for synchronous logic.

            Mock Response File: post_limit_order_cancel.json
        """
        user_public_key = ""
        fee_payer_public_key = ""
        orders = []

        # load mock response
        mock_file_name = "post_limit_order_cancel"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, PostLimitOrderCancelResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
        else:
            # find open orders
            open_orders = self.jupiter.client.get_limit_order_open(input_token = JUP, output_token = BONK)
            order = open_orders.orders[0]

            # set inputs
            user_public_key = order.account.maker
            fee_payer_public_key = order.account.maker
            orders = [order.public_key]

        # create input
        body = PostLimitOrderCancelBody(
            user_public_key = user_public_key,
            fee_payer_public_key = fee_payer_public_key,
            orders = orders
        )

        # execute request
        response = self.jupiter.client.post_limit_order_cancel(body)

        # actual test
        assert isinstance(response, PostLimitOrderCancelResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_post_limit_order_cancel_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint POST "Limit Order Cancel" 
            for asynchronous logic.

            Mock Response File: post_limit_order_cancel.json
        """
        user_public_key = ""
        fee_payer_public_key = ""
        orders = []

        async with self.jupiter.async_client as client:
            # load mock response
            mock_file_name = "post_limit_order_cancel"
            if config.mock_response or config.jupiter.mock_response:
                mock_response = self.mocker.load_mock_response(mock_file_name, PostLimitOrderCancelResponse)
                mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            else:
                # find open orders
                open_orders = await client.get_limit_order_open(input_token = JUP, output_token = BONK)
                order = open_orders.orders[0]

                # set inputs
                user_public_key = order.account.maker
                fee_payer_public_key = order.account.maker
                orders = [order.public_key]

            # create input
            body = PostLimitOrderCancelBody(
                user_public_key = user_public_key,
                fee_payer_public_key = fee_payer_public_key,
                orders = orders
            )

            # execute request
            response = await client.post_limit_order_cancel(body)

        # actual test
        assert isinstance(response, PostLimitOrderCancelResponse)

    def test_post_limit_order_cancel_invalid_request_sync(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order Cancel" 
            when an invalid field is provided in the body for synchronous logic.
        """
        body = PostLimitOrderCancelBody(
            user_public_key = "",
            fee_payer_public_key = "",
            orders = []
        )
        with pytest.raises(JupiterException):
            self.jupiter.client.post_limit_order_cancel(body)

    @pytest.mark.asyncio
    async def test_post_limit_order_cancel_invalid_request_async(self) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order Cancel" 
            when an invalid field is provided in the body for asynchronous logic.
        """
        body = PostLimitOrderCancelBody(
            user_public_key = "",
            fee_payer_public_key = "",
            orders = []
        )
        with pytest.raises(JupiterException):
            async with self.jupiter.async_client as client:
                await client.post_limit_order_cancel(body)

    def test_get_limit_order_open_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - Open" 
            for synchronous logic.

            Mock Response File: get_limit_order_open.json
        """

        # load mock response
        mock_file_name = "get_limit_order_open"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderOpenResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_limit_order_open(input_token = JUP, output_token = BONK)

        # actual test
        assert isinstance(response, GetLimitOrderOpenResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            response.orders = response.orders[0:5]
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_limit_order_open_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - Open" 
            for asynchronous logic.

            Mock Response File: get_limit_order_open.json
        """

        # load mock response
        mock_file_name = "get_limit_order_open"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderOpenResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_limit_order_open(input_token = JUP, output_token = BONK)

        # actual test
        assert isinstance(response, GetLimitOrderOpenResponse)

    def test_get_limit_order_history_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - History" 
            for synchronous logic.

            Mock Response File: get_limit_order_history.json
        """

        # load mock response
        mock_file_name = "get_limit_order_history"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderHistoryResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_limit_order_history(
            wallet = "Hq9YQ2sz6A318tdNbFWMpML6AjWX3wDTLPVx26m719qG",
            take = 1
        )

        # actual test
        assert isinstance(response, GetLimitOrderHistoryResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            response.orders = [response.orders[0]]
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_limit_order_history_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - History" 
            for asynchronous logic.

            Mock Response File: get_limit_order_history.json
        """

        # load mock response
        mock_file_name = "get_limit_order_history"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderHistoryResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_limit_order_history(
                wallet = "Hq9YQ2sz6A318tdNbFWMpML6AjWX3wDTLPVx26m719qG",
                take = 1
            )

        # actual test
        assert isinstance(response, GetLimitOrderHistoryResponse)

    def test_get_limit_order_trade_history_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - Trade History" 
            for synchronous logic.

            Mock Response File: get_limit_order_trade_history.json
        """

        # load mock response
        mock_file_name = "get_limit_order_trade_history"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderTradeHistoryResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.jupiter.client.get_limit_order_trade_history(
            wallet = "Hq9YQ2sz6A318tdNbFWMpML6AjWX3wDTLPVx26m719qG",
            take = 1
        )

        # actual test
        assert isinstance(response, GetLimitOrderTradeHistoryResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.jupiter.mock_response:
            response.orders = [response.orders[0]]
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_limit_order_trade_history_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Limit Order - Trade History" 
            for asynchronous logic.

            Mock Response File: get_limit_order_trade_history.json
        """

        # load mock response
        mock_file_name = "get_limit_order_trade_history"
        if config.mock_response or config.jupiter.mock_response:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetLimitOrderTradeHistoryResponse)

            # response content to be adjusted
            content = str(mock_response.json()["orders"]).replace("'", '"').encode()
            mock_response._content = content

            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.jupiter.async_client as client:
            response = await client.get_limit_order_trade_history(
                wallet = "Hq9YQ2sz6A318tdNbFWMpML6AjWX3wDTLPVx26m719qG",
                take = 1
            )

        # actual test
        assert isinstance(response, GetLimitOrderTradeHistoryResponse)


from typing import Any
from enum import Enum

from ..core.exception import ParamUnknownError

class CyholeParam(Enum):
    """
        Generic Enum representing a param for an API endpoint
    """
    @classmethod
    def check(cls, value: Any) -> None:
        """
            Function used to check the consistency of an input API 
            parameter (param) belonging to the value list (enum.Enum).

            The function will raise a ParamUnknownError if the value
            does not belong to the Enum.

            Parameters:
                value: value to check.
        """
        if value not in cls:
            raise ParamUnknownError(value, cls)
        return

class RequestType(CyholeParam):
    """
        Enum class for RequestType
    """
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


from ..core.exception import CyholeException

class BirdeyeException(CyholeException):
    pass

class BirdeyeAuthorisationError(BirdeyeException):
    def __init__(self, description: str = "The API Key is not authorised to use this endpoint."):
        super().__init__(description)

class BirdeyeTimeRangeError(BirdeyeException):
    pass


import os
from typing import Coroutine, Literal, overload
from datetime import datetime

from ..core.param import RequestType
from ..core.interaction import Interaction
from ..core.exception import MissingAPIKeyError
from ..birdeye.client import BirdeyeClient, BirdeyeAsyncClient
from ..birdeye.exception import BirdeyeTimeRangeError
from ..birdeye.param import (
    BirdeyeChain,
    BirdeyeOrder,
    BirdeyeSort,
    BirdeyeTimeFrame,
    BirdeyeTradeType,
    BirdeyeAddressType
)
from ..birdeye.schema import (
    GetTokenListResponse,
    GetTokenSecurityResponse,
    GetTokenCreationInfoResponse,
    GetTokenOverviewResponse,
    GetPriceResponse,
    GetPriceMultipleResponse,
    GetPriceHistoricalResponse,
    GetTradesTokenResponse,
    GetTradesPairResponse,
    GetOHLCVTokenPairResponse,
    GetOHLCVBaseQuoteResponse,
    GetWalletSupportedNetworksResponse
)

class Birdeye(Interaction):
    """
        Class used to connect [https://birdeye.so](https://birdeye.so) API.
        To have access Birdeye API (public or private) is required to have a valid API key.

        Check [https://docs.birdeye.so](https://docs.birdeye.so) for all the details on the available endpoints.

        !!! info
            If the API key is not provided during the object creation, then it is automatically 
            retrieved from ENV variable **BIRDEYE_API_KEY**.

        Parameters:
            api_key: specify the API key to use for the connection.
            chain: identifier of the chain to use in all the requests.
                The supported chains are available on [`BirdeyeChain`][cyhole.birdeye.param.BirdeyeChain].
                Import them from the library to use the correct identifier.

        **Example**
        ```python
        import asyncio
        from cyhole.birdeye import Birdeye

        birdeye = Birdeye()

        # Get current token list on Solana
        # synchronous
        response = birdeye.client.get_token_list()
        print(f"Currently listed {len(response.data.tokens)} tokens on Solana")

        # asynchronous
        async def main() -> None:
            async with birdeye.async_client as client:
                response = await client.get_token_list()
                print(f"Currently listed {len(response.data.tokens)} tokens on Solana")

        asyncio.run(main())
        ```

        Raises:
            MissingAPIKeyError: if no API Key was available during the object creation.
    """
    def __init__(self, api_key: str | None = None, chain: str = BirdeyeChain.SOLANA.value) -> None:

        # set API
        self.api_key = api_key if api_key is not None else os.environ.get("BIRDEYE_API_KEY")
        if self.api_key is None:
            raise MissingAPIKeyError("no API key is provided during object's creation.")

        # input check

        # headers setup
        headers = {
            "X-API-KEY": self.api_key,
            "x-chain": chain
        }
        super().__init__(headers)
        self.headers: dict[str, str]

        # clients
        self.client = BirdeyeClient(self, headers = headers)
        self.async_client = BirdeyeAsyncClient(self, headers = headers)

        # API urls
        self.url_api_public = "https://public-api.birdeye.so/defi/"
        self.url_api_private = "https://public-api.birdeye.so/defi/"
        self.url_api_private_wallet = "https://public-api.birdeye.so/v1/wallet"
        return

    @overload
    def _get_token_list(
        self,
        sync: Literal[True],
        sort_by: str = BirdeyeSort.SORT_V24HUSD.value,
        order_by: str = BirdeyeOrder.DESCENDING.value,
        offset: int | None = None,
        limit: int | None = None
    ) -> GetTokenListResponse: ...

    @overload
    def _get_token_list(
        self,
        sync: Literal[False],
        sort_by: str = BirdeyeSort.SORT_V24HUSD.value,
        order_by: str = BirdeyeOrder.DESCENDING.value,
        offset: int | None = None,
        limit: int | None = None
    ) -> Coroutine[None, None, GetTokenListResponse]: ...

    def _get_token_list(
        self,
        sync: bool,
        sort_by: str = BirdeyeSort.SORT_V24HUSD.value,
        order_by: str = BirdeyeOrder.DESCENDING.value,
        offset: int | None = None,
        limit: int | None = None
    ) -> GetTokenListResponse | Coroutine[None, None, GetTokenListResponse]:
        """
            This function refers to the **PUBLIC** API endpoint **[Token - List](https://docs.birdeye.so/reference/get_defi-tokenlist)** and is used 
            to get the list of Birdeye tokens according on a specific chain.

            Parameters:
                sort_by: define the type of sorting to apply in the
                    extraction; e.g. USD volume in the last 24h.
                    The sorting types are available on [`BirdeyeSort`][cyhole.birdeye.param.BirdeyeSort].
                    Import them from the library to use the correct identifier.
                order_by: define the type of ordering to apply in the 
                    extraction; e.g. ascending or descending.
                    The sorting types are available on [`BirdeyeOrder`][cyhole.birdeye.param.BirdeyeOrder].
                    Import them from the library to use the correct identifier.
                offset: offset to apply in the extraction.
                limit: limit the number of returned records in the extraction.

            Returns:
                list of tokens returned by birdeye.so

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeSort.check(sort_by)
        BirdeyeOrder.check(order_by)

        # set params
        url = self.url_api_public + "tokenlist"
        params = {
            "sort_by" : sort_by,
            "sort_type" : order_by,
            "offset" : offset,
            "limit": limit
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTokenListResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTokenListResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_token_creation_info(
        self,
        sync: Literal[True],
        address: str
    ) -> GetTokenCreationInfoResponse: ...

    @overload
    def _get_token_creation_info(
        self,
        sync: Literal[False],
        address: str
    ) -> Coroutine[None, None, GetTokenCreationInfoResponse]: ...

    def _get_token_creation_info(
        self,
        sync: bool,
        address: str
    ) -> GetTokenCreationInfoResponse | Coroutine[None, None, GetTokenCreationInfoResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Token - Creation Token Info](https://docs.birdeye.so/reference/get_defi-token-creation-info)** and is used 
            to get the current price of a token according on a specific chain on Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
            
            Returns:
                token's creation information.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # set params
        url = self.url_api_public + "token_creation_info"
        params = {
            "address" : address
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTokenCreationInfoResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTokenCreationInfoResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_token_security(
        self,
        sync: Literal[True],
        address: str
    ) -> GetTokenSecurityResponse: ...

    @overload
    def _get_token_security(
        self,
        sync: Literal[False],
        address: str
    ) -> Coroutine[None, None, GetTokenSecurityResponse]: ...

    def _get_token_security(
        self,
        sync: bool,
        address: str
    ) -> GetTokenSecurityResponse | Coroutine[None, None, GetTokenSecurityResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Token - Security](https://docs.birdeye.so/reference/get_defi-token-security)** and is used 
            to get the useful information related to the security of a token  on a specific
            chain calculated by Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
            
            Returns:
                token's security information.
                    Observe that the content of `data` value depends on the selected chain.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # set params
        url = self.url_api_public + "token_security"
        params = {
            "address" : address
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTokenSecurityResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTokenSecurityResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_token_overview(
        self,
        sync: Literal[True],
        address: str
    ) -> GetTokenOverviewResponse: ...

    @overload
    def _get_token_overview(
        self,
        sync: Literal[False],
        address: str
    ) -> Coroutine[None, None, GetTokenOverviewResponse]: ...

    def _get_token_overview(
        self,
        sync: bool,
        address: str
    ) -> GetTokenOverviewResponse | Coroutine[None, None, GetTokenOverviewResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Token - Overview](https://docs.birdeye.so/reference/get_defi-token-overview)** and is used 
            to get all kind of information (token/mint/creator adresses, high level statistics, ...)
            of a token on a specific chain calculated by Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
            
            Returns:
                token's information.
                    Observe that the content of `data` value depends on the selected chain.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # set params
        url = self.url_api_public + "token_overview"
        params = {
            "address" : address
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTokenOverviewResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTokenOverviewResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_price(
        self,
        sync: Literal[True],
        address: str,
        include_liquidity: bool | None = None
    ) -> GetPriceResponse: ...

    @overload
    def _get_price(
        self,
        sync: Literal[False],
        address: str,
        include_liquidity: bool | None = None
    ) -> Coroutine[None, None, GetPriceResponse]: ...

    def _get_price(
        self,
        sync: bool,
        address: str,
        include_liquidity: bool | None = None
    ) -> GetPriceResponse | Coroutine[None, None, GetPriceResponse]:
        """
            This function refers to the **PUBLIC** API endpoint **[Price](https://docs.birdeye.so/reference/get_defi-price)** and is used 
            to get the current price of a token according on a specific chain on Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
                include_liquidity: include the current liquidity of the token.
                    Default Value: `None` (`False`)

            Returns:
                token's price returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """        # set params
        url = self.url_api_public + "price"
        params = {
            "address" : address,
            "include_liquidity" : str(include_liquidity).lower() if include_liquidity else None
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetPriceResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetPriceResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_price_multiple(
        self,
        sync: Literal[True],
        list_address: list[str],
        include_liquidity: bool | None = None
    ) -> GetPriceMultipleResponse: ...

    @overload
    def _get_price_multiple(
        self,
        sync: Literal[False],
        list_address: list[str],
        include_liquidity: bool | None = None
    ) -> Coroutine[None, None, GetPriceMultipleResponse]: ...

    def _get_price_multiple(
        self,
        sync: bool,
        list_address: list[str],
        include_liquidity: bool | None = None
    ) -> GetPriceMultipleResponse | Coroutine[None, None, GetPriceMultipleResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Price - Multiple](https://docs.birdeye.so/reference/get_defi-multi-price)** and is used 
            to get the current price of multeple tokens on a specific chain on Birdeye.

            Parameters:
                list_address: CA of the tokens to search on the chain.
                include_liquidity: include the current liquidity of the token.
                    Default Value: `None` (`False`)

            Returns:
                list of tokens returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # set params
        url = self.url_api_public + "multi_price"
        params = {
            "list_address" : ",".join(list_address),
            "include_liquidity" : str(include_liquidity).lower() if include_liquidity else None
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetPriceMultipleResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetPriceMultipleResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_price_historical(
        self,
        sync: Literal[True],
        address: str,
        address_type: str,
        timeframe: str,
        dt_from: datetime,
        dt_to: datetime | None = None
    ) -> GetPriceHistoricalResponse: ...

    @overload
    def _get_price_historical(
        self,
        sync: Literal[False],
        address: str,
        address_type: str,
        timeframe: str,
        dt_from: datetime,
        dt_to: datetime | None = None
    ) -> Coroutine[None, None, GetPriceHistoricalResponse]: ...

    def _get_price_historical(
        self,
        sync: bool,
        address: str,
        address_type: str,
        timeframe: str,
        dt_from: datetime,
        dt_to: datetime | None = None
    ) -> GetPriceHistoricalResponse | Coroutine[None, None, GetPriceHistoricalResponse]:
        """
            This function refers to the **PUBLIC** API endpoint **[Price - Historical](https://docs.birdeye.so/reference/get_defi-history-price)** and is used 
            to get the history of prices of a token according on a specific chain on Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
                address_type: the type of address involved in the extraction.
                    The supported chains are available on [`BirdeyeAddressType`][cyhole.birdeye.param.BirdeyeAddressType].
                    Import them from the library to use the correct identifier.
                timeframe: the type of timeframe involved in the extraction.
                    The timeframe is used to define intervall between a measure and the next one.
                    The supported chains are available on [`BirdeyeTimeFrame`][cyhole.birdeye.param.BirdeyeTimeFrame].
                    Import them from the library to use the correct identifier.
                dt_from: beginning time to take take price data.
                dt_to: end time to take take price data.
                    It should be `dt_from` < `dt_to`.
                    If not ptovided (None), the current time is used.

            Returns:
                list of prices returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeAddressType.check(address_type)
        BirdeyeTimeFrame.check(timeframe)

        # set default
        if dt_to is None:
            dt_to = datetime.now()

        # check consistency
        if dt_from > dt_to:
            raise BirdeyeTimeRangeError("Inconsistent timewindow provided: 'dt_from' > 'dt_to'")

        # set params
        url = self.url_api_public + "history_price"
        params = {
            "address" : address,
            "address_type" : address_type,
            "type" : timeframe,
            "time_from" : int(dt_from.timestamp()),
            "time_to" : int(dt_to.timestamp())
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetPriceHistoricalResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetPriceHistoricalResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_trades_token(
            self,
            sync: Literal[True],
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> GetTradesTokenResponse: ...

    @overload
    def _get_trades_token(
            self,
            sync: Literal[False],
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> Coroutine[None, None, GetTradesTokenResponse]: ...

    def _get_trades_token(
            self,
            sync: bool,
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> GetTradesTokenResponse | Coroutine[None, None, GetTradesTokenResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Trades - Token](https://docs.birdeye.so/reference/get_defi-txs-token)** and is used 
            to get the associated trades of a token according on a specific chain on Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
                trade_type: the type of transactions to extract.
                    The supported chains are available on [`BirdeyeTradeType`][cyhole.birdeye.param.BirdeyeTradeType].
                    Import them from the library to use the correct identifier.
                offset: offset to apply in the extraction.
                limit: limit the number of returned records in the extraction.

            Returns:
                list of prices returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeTradeType.check(trade_type)

        # set params
        url = self.url_api_private + "txs/token"
        params = {
            "address" : address,
            "tx_type" : trade_type,
            "offset" : offset,
            "limit": limit
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTradesTokenResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTradesTokenResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_trades_pair(
            self,
            sync: Literal[True],
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            order_by: str = BirdeyeOrder.DESCENDING.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> GetTradesPairResponse: ...

    @overload
    def _get_trades_pair(
            self,
            sync: Literal[False],
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            order_by: str = BirdeyeOrder.DESCENDING.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> Coroutine[None, None, GetTradesPairResponse]: ...

    def _get_trades_pair(
            self,
            sync: bool,
            address: str,
            trade_type: str = BirdeyeTradeType.SWAP.value,
            order_by: str = BirdeyeOrder.DESCENDING.value,
            offset: int | None = None,
            limit: int | None = None
    ) -> GetTradesPairResponse | Coroutine[None, None, GetTradesPairResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Trades - Pair](https://docs.birdeye.so/reference/get_defi-txs-pair)** and is used 
            to get the associated trades of a tokens pair according on a specific chain on Birdeye. 
            Use the 'Trades - Token' endpoint to retrieve the trades associated to a specific token.

            Parameters:
                address: CA of the token to search on the chain.
                trade_type: the type of transactions to extract.
                    The supported chains are available on [`BirdeyeTradeType`][cyhole.birdeye.param.BirdeyeTradeType].
                    Import them from the library to use the correct identifier.
                order_by: define the type of ordering to apply in the 
                    extraction; e.g. ascending or descending.
                    The sorting types are available on [`BirdeyeOrder`][cyhole.birdeye.param.BirdeyeOrder].
                    Import them from the library to use the correct identifier.
                offset: offset to apply in the extraction.
                limit: limit the number of returned records in the extraction.

            Returns:
                list of prices returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeOrder.check(order_by)
        BirdeyeTradeType.check(trade_type)

        # set params
        url = self.url_api_private + "txs/pair"
        params = {
            "address" : address,
            "tx_type" : trade_type,
            "sort_type": order_by,
            "offset" : offset,
            "limit": limit
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTradesPairResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTradesPairResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_ohlcv(
            self,
            sync: Literal[True],
            address: str,
            address_type: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> GetOHLCVTokenPairResponse: ...

    @overload
    def _get_ohlcv(
            self,
            sync: Literal[False],
            address: str,
            address_type: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> Coroutine[None, None, GetOHLCVTokenPairResponse]: ...

    def _get_ohlcv(
            self,
            sync: bool,
            address: str,
            address_type: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> GetOHLCVTokenPairResponse | Coroutine[None, None, GetOHLCVTokenPairResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[OHLCV - Token/Pair](https://docs.birdeye.so/reference/get_defi-ohlcv)** and is used to get the 
            Open, High, Low, Close, and Volume (OHLCV) data for a specific token/pair on a chain on Birdeye.

            Parameters:
                address: CA of the token to search on the chain.
                address_type: the type of address involved in the extraction (token/pair).
                    The supported chains are available on [`BirdeyeAddressType`][cyhole.birdeye.param.BirdeyeAddressType].
                    Import them from the library to use the correct identifier.
                timeframe: the type of timeframe involved in the extraction.
                    The timeframe is used to define intervall between a measure and the next one.
                    The supported chains are available on [`BirdeyeTimeFrame`][cyhole.birdeye.param.BirdeyeTimeFrame].
                    Import them from the library to use the correct identifier.
                dt_from: beginning time to take take price data.
                dt_to: end time to take take price data.
                    It should be `dt_from` < `dt_to`.
                    If not ptovided (None), the current time is used.

            Returns:
                list of prices returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeAddressType.check(address_type)
        BirdeyeTimeFrame.check(timeframe)

        # set default
        if dt_to is None:
            dt_to = datetime.now()

        # check consistency
        if dt_from > dt_to:
            raise BirdeyeTimeRangeError("Inconsistent timewindow provided: 'dt_from' > 'dt_to'")

        # set params
        url = self.url_api_public + "ohlcv"
        if address_type == BirdeyeAddressType.PAIR.value:
            url = url + "/" + BirdeyeAddressType.PAIR.value
        params = {
            "address" : address,
            "type" : timeframe,
            "time_from" : int(dt_from.timestamp()),
            "time_to" : int(dt_to.timestamp())
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetOHLCVTokenPairResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetOHLCVTokenPairResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_ohlcv_base_quote(
            self,
            sync: Literal[True],
            base_address: str,
            quote_address: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> GetOHLCVBaseQuoteResponse: ...

    @overload
    def _get_ohlcv_base_quote(
            self,
            sync: Literal[False],
            base_address: str,
            quote_address: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> Coroutine[None, None, GetOHLCVBaseQuoteResponse]: ...

    def _get_ohlcv_base_quote(
            self,
            sync: bool,
            base_address: str,
            quote_address: str,
            timeframe: str,
            dt_from: datetime,
            dt_to: datetime | None = None,
            chain: str = BirdeyeChain.SOLANA.value
    ) -> GetOHLCVBaseQuoteResponse | Coroutine[None, None, GetOHLCVBaseQuoteResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[OHLCV - Base/Quote](https://docs.birdeye.so/reference/get_defi-ohlcv-base-quote)** and is used to get the 
            Open, High, Low, Close, and Volume (OHLCV) data for a specific base/quote combination 
            on a chain on Birdeye.

            Parameters:
                base_address: CA of the token to search on the chain.
                quote_address: CA of the token to search on the chain.
                timeframe: the type of timeframe involved in the extraction.
                    The timeframe is used to define intervall between a measure and the next one.
                    The supported chains are available on [`BirdeyeTimeFrame`][cyhole.birdeye.param.BirdeyeTimeFrame].
                    Import them from the library to use the correct identifier.
                dt_from: beginning time to take take price data.
                dt_to: end time to take take price data.
                    It should be `dt_from` < `dt_to`.
                    If not ptovided (None), the current time is used.
            Returns:
                list of prices returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
                ParamUnknownError: if one of the input parameter belonging to the value list is aligned to it.
        """
        # check param consistency
        BirdeyeTimeFrame.check(timeframe)

        # set default
        if dt_to is None:
            dt_to = datetime.now()

        # check consistency
        if dt_from > dt_to:
            raise BirdeyeTimeRangeError("Inconsistent timewindow provided: 'dt_from' > 'dt_to'")

        # set params
        url = self.url_api_public + "ohlcv/base_quote"
        params = {
            "base_address" : base_address,
            "quote_address" : quote_address,
            "type" : timeframe,
            "time_from" : int(dt_from.timestamp()),
            "time_to" : int(dt_to.timestamp())
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetOHLCVBaseQuoteResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetOHLCVBaseQuoteResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_wallet_supported_networks(self, sync: Literal[True]) -> GetWalletSupportedNetworksResponse: ...

    @overload
    def _get_wallet_supported_networks(self, sync: Literal[False]) -> Coroutine[None, None, GetWalletSupportedNetworksResponse]: ...

    def _get_wallet_supported_networks(self, sync: bool) -> GetWalletSupportedNetworksResponse | Coroutine[None, None, GetWalletSupportedNetworksResponse]:
        """
            This function refers to the **PRIVATE** API endpoint **[Wallet - Supported Networks](https://docs.birdeye.so/reference/get_v1-wallet-list-supported-chain)** and 
            it is used to get the list of supported chains on Birdeye.

            Returns:
                list of chains returned by birdeye.so.

            Raises:
                BirdeyeAuthorisationError: if the API key provided does not give access to related endpoint.
        """
        url = self.url_api_private_wallet + "/list_supported_chain"

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url)
            return GetWalletSupportedNetworksResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url)
                return GetWalletSupportedNetworksResponse(**content_raw.json())
            return async_request()